import msvcrt
import ctypes
import os
import ctypes
from typing import Any, Union, IO
from os import PathLike
from pathlib import Path
from types import MethodType
from ._iotype import _IOType


def open(
    filename: Union[str, PathLike[str]], mode: str = "r", *args: Any, **kwargs: Any
) -> _IOType:
    """
    Opens a file with specified mode using Windows API to handle various locks.

    This function works similarly to the built-in `open` but can handle files locked
    by programs such as OneDrive. It bypasses locks for reading using the correct
    combination of flags on the Windows API.

    Args:
        filename (str): The path to the file to be opened.
        mode (str, optional): The mode in which the file is opened. Supports 'r', 'w', 'rw',
            'rb', 'wb', 'a', and 'ab'. Defaults to 'r'.
        *args: Variable list of arguments passed to os.fdopen().
        **kwargs: Arbitrary keyword arguments passed to os.fdopen().

    Returns:
        file: A file object.

    Raises:
        ValueError: If the provided mode is not supported.
        FileNotFoundError: If the file does not exist.
        PermissionError: If permission is denied to access the file.
        FileExistsError: If the file already exists and mode is set to create.
        IsADirectoryError: If the provided filename is a directory.
        OSError: For generic OS errors and errors thrown by os.fdopen.
        ctypes.WinError: For other Windows API-related errors.

    Example:
        >>> with open("locked_file.txt", "r") as f:
        >>>     content = f.read()

    Note:
        This function is designed for Windows and uses the Windows API. It is not cross-platform.

    """
    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x40000000
    FILE_SHARE_READ = 1
    FILE_SHARE_WRITE = 2
    FILE_SHARE_DELETE = 4

    OPEN_EXISTING = 3
    CREATE_ALWAYS = 2
    OPEN_ALWAYS = 4
    FILE_END = 2

    access_flags = {
        "r": GENERIC_READ,
        "w": GENERIC_WRITE,
        "rw": GENERIC_READ | GENERIC_WRITE,
        "rb": GENERIC_READ,
        "wb": GENERIC_WRITE,
        "a": GENERIC_WRITE,  # for 'append' mode
        "ab": GENERIC_WRITE,  # for 'append' binary mode
    }

    dispositions = {
        "r": OPEN_EXISTING,
        "w": CREATE_ALWAYS,
        "rw": OPEN_ALWAYS,
        "rb": OPEN_EXISTING,
        "wb": CREATE_ALWAYS,
        "a": OPEN_ALWAYS,  # for 'append' mode
        "ab": OPEN_ALWAYS,  # for 'append' binary mode
    }

    # Ignore to avoid mypy error on Linux machines
    CreateFileW = ctypes.windll.kernel32.CreateFileW  # type: ignore

    access_mode = access_flags.get(mode)
    if access_mode is None:
        raise ValueError(
            f"Invalid mode '{mode}', only 'r', 'w', 'rw', 'rb', 'wb', 'a', and 'ab' are supported"
        )

    disposition = dispositions.get(mode)
    if disposition is None:
        raise ValueError(
            f"Invalid mode '{mode}', only 'r', 'w', 'rw', 'rb', 'wb', 'a', and 'ab' are supported"
        )

    hfile = CreateFileW(
        str(Path(filename)),
        access_mode,
        FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
        None,
        disposition,
        0,
        None,
    )

    UNKNOWN_ERROR = 0
    INVALID_HANDLE_VALUE = -1
    ERROR_FILE_NOT_FOUND = 2
    ERROR_ACCESS_DENIED = 5
    ERROR_INVALID_PARAMETER = 87
    ERROR_FILE_EXISTS = 80
    ERROR_ALREADY_EXISTS = 183
    ERROR_DIRECTORY = 267

    if hfile == INVALID_HANDLE_VALUE:
        # Ignore to avoid mypy error on Linux machines
        error_code = ctypes.GetLastError()  # type: ignore

        if error_code == ERROR_FILE_NOT_FOUND:
            raise FileNotFoundError(f"No such file or directory: '{filename}'")
        elif error_code == ERROR_ACCESS_DENIED:
            raise PermissionError("Permission denied: '{}'".format(filename))
        elif error_code == ERROR_INVALID_PARAMETER:
            raise ValueError("Invalid parameter")
        elif error_code == ERROR_FILE_EXISTS or error_code == ERROR_ALREADY_EXISTS:
            raise FileExistsError("File already exists: '{}'".format(filename))
        elif error_code == ERROR_DIRECTORY:
            raise IsADirectoryError("Is a directory: '{}'".format(filename))
        elif error_code == UNKNOWN_ERROR:
            raise OSError("Unknown error")
        else:
            # Ignore to avoid mypy error on Linux machines
            raise ctypes.WinError()  # type: ignore

    try:
        # for 'append' mode, you'd also need to move the file pointer to the end
        if mode in {"a", "ab"}:
            # Ignore to avoid mypy error on Linux machines
            ctypes.windll.kernel32.SetFilePointer(hfile, 0, None, FILE_END)  # type: ignore

        # Convert the Windows handle into a C runtime file descriptor
        # Ignore to avoid mypy error on Linux machines
        fd = msvcrt.open_osfhandle(hfile, os.O_BINARY if "b" in mode else os.O_TEXT)  # type: ignore

        # Create a Python file object from the file descriptor
        file = os.fdopen(fd, mode, *args, **kwargs)
    except:
        # Close the handle if an exception occurred
        # Ignore to avoid mypy error on Linux machines
        ctypes.windll.kernel32.CloseHandle(hfile)  # type: ignore
        raise

    # Monkeypatch the file object to include hfile as a closure and
    # apply `CloseHandle(hfile)` when closing the file
    original_exit = file.__exit__

    def __exit__(self: IO[Any], *args: Any) -> None:
        original_exit(*args)
        # Ignore to avoid mypy error on Linux machines
        ctypes.windll.kernel32.CloseHandle(hfile)  # type: ignore

    setattr(
        file,
        "__exit__",
        MethodType(__exit__, file),
    )

    return file



import unittest
import random
import string
import locate
from pathlib import Path
import os
import tempfile
import uuid
import time
import psutil
import shutil
import pythoncom
import win32com.client
import win32process
import win32com
import ctypes
import multiprocessing

with locate.prepend_sys_path(".."):
    from winunlock import open as winunlock_open
    from winunlock import monkeypatch_open


# Note that this type of lock is still too strogn for our reading function
def lock_file(filename):
    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x40000000
    OPEN_EXISTING = 3

    CreateFile = ctypes.windll.kernel32.CreateFileW

    hfile = CreateFile(
        filename,
        GENERIC_READ | GENERIC_WRITE,  # open for reading and writing
        0,  # no sharing
        None,
        OPEN_EXISTING,
        0,
        None,
    )
    if hfile == -1:
        raise ctypes.WinError()

    # Keep the process running and holding the lock
    while True:
        time.sleep(1)


class GenerateLockFile:
    def __init__(self):
        self.temp_filename_path = None
        self.process = None

    def __enter__(self):
        # create a temporary file
        self.temp_filename_path = Path(tempfile.gettempdir()) / f"{uuid.uuid4()}.txt"
        with open(self.temp_filename_path, "w") as temp_file:
            temp_file.write("This is a temporary file.")

        # open the temporary file in a new lock_file process
        self.process = multiprocessing.Process(
            target=lock_file, args=(str(self.temp_filename_path),)
        )
        self.process.start()

        # wait for lock_file to create the lock
        i = 0
        while (i := i + 1) != 50 and self.process.is_alive():
            time.sleep(0.1)

        if not self.process.is_alive():
            raise Exception("Could not create lock file")

        # return the lock file path
        return self.temp_filename_path

    def __exit__(self, *args):
        # terminate the lock_file process
        self.process.terminate()

        # delete the temporary file
        i = 0
        while (i := i + 1) != 50:
            try:
                os.remove(self.temp_filename_path)
                break
            except PermissionError:
                time.sleep(0.1)

        if i == 50:
            raise Exception("Could not delete temporary file")


class GenerateLockFileUsingExcel:
    def __init__(self):
        self.excel_path = locate.this_dir() / "example.xlsx"
        self.temp_excel_path = None
        self.process = None

    def __enter__(self):
        # copy the Excel file to a temporary location
        temp_excel_name = f"{uuid.uuid4()}.xlsx"
        self.temp_excel_path = Path(tempfile.gettempdir()) / temp_excel_name
        shutil.copy(self.excel_path, self.temp_excel_path)

        # open the temporary Excel file in a new Excel process
        pythoncom.CoInitialize()  # Initialize the COM library
        excel = win32com.client.DispatchEx(
            "Excel.Application"
        )  # creates a new instance of Excel
        excel.Visible = 0  # makes Excel hidden
        excel.Workbooks.Open(str(self.temp_excel_path))  # opens the workbook

        # Get the main thread ID and process ID of the Excel instance
        hndl = excel.Hwnd
        _, pid = win32process.GetWindowThreadProcessId(hndl)
        self.process = psutil.Process(pid)

        # wait for Excel to create the lock file
        lock_file_name = f"~${self.temp_excel_path.stem}.xlsx"
        lock_file_path = Path(tempfile.gettempdir()) / lock_file_name

        i = 0
        while (i := i + 1) != 50 and not lock_file_path.exists():
            time.sleep(0.1)

        if not lock_file_path.exists():
            raise Exception("Could not create lock file")

        # return the lock file path
        return lock_file_path

    def __exit__(self, *args):
        # close the Excel process
        p = psutil.Process(self.process.pid)
        p.terminate()

        # wait for Excel to remove the lock file
        lock_file_name = f"~${self.temp_excel_path.stem}.xlsx"
        lock_file_path = Path(tempfile.gettempdir()) / lock_file_name

        i = 0
        while (i := i + 1) != 50 and lock_file_path.exists():
            time.sleep(0.1)

        if lock_file_path.exists():
            raise Exception("Could not exit Excel process")

        # delete the temporary Excel file
        os.remove(self.temp_excel_path)


class TestOpenLockedWithExcelLock(unittest.TestCase):
    def setUp(self):
        self.xl = GenerateLockFileUsingExcel()
        self.fname = self.xl.__enter__()

    def tearDown(self):
        self.xl.__exit__(None, None, None)

    def test_read_text_file_with_open(self):
        # with the correct error
        with self.assertRaises(PermissionError):
            with open(
                self.fname,
                "r",
            ) as f:
                pass

    def test_read_text_file_with_winunlock_open(self):
        # success without error
        with winunlock_open(
            self.fname,
            "r",
        ) as f:
            f.read()

    def test_monkeypatch_open(self):
        # success without error
        with monkeypatch_open():
            with open(
                self.fname,
                "r",
            ) as f:
                f.read()


class TestOpenLockedWithLock(unittest.TestCase):
    def setUp(self):
        self.glf = GenerateLockFile()
        self.fname = self.glf.__enter__()

    def tearDown(self):
        self.glf.__exit__(None, None, None)

    def test_read_text_file_with_open(self):
        # with the correct error
        with self.assertRaises(PermissionError):
            with open(
                self.fname,
                "r",
            ) as f:
                pass

    # TODO: find a way to read this locked file
    """
    def test_read_text_file_with_winunlock_open(self):
        # success without error
        with winunlock_open(
            self.fname,
            "r",
        ) as f:
            f.read()
    """


class TestOpenLocked(unittest.TestCase):
    def random_string(self, length):
        return "".join(random.choice(string.ascii_letters) for _ in range(length))

    def test_read_text_file(self):
        # create a temp file with some text
        filename = tempfile.mktemp()
        with open(filename, "w") as f:
            f.write("hello world")

        # try reading it with winunlock_open
        with winunlock_open(filename, "r") as f:
            data = f.read()

        self.assertEqual(data, "hello world")

        # clean up
        os.remove(filename)

    def test_read_binary_file(self):
        # create a temp file with some bytes
        filename = tempfile.mktemp()
        with open(filename, "wb") as f:
            f.write(b"\x00\x01\x02\x03\x04")

        # try reading it with winunlock_open
        with winunlock_open(filename, "rb") as f:
            data = f.read()

        self.assertEqual(data, b"\x00\x01\x02\x03\x04")

        # clean up
        os.remove(filename)

    def test_write_text_file(self):
        filename = tempfile.mktemp()

        # try writing to it with winunlock_open
        with winunlock_open(filename, "w") as f:
            f.write("hello world")

        # check that the file contains the correct data
        with open(filename, "r") as f:
            data = f.read()

        self.assertEqual(data, "hello world")

        # clean up
        os.remove(filename)

    def test_write_binary_file(self):
        filename = tempfile.mktemp()

        # try writing to it with winunlock_open
        with winunlock_open(filename, "wb") as f:
            f.write(b"\x00\x01\x02\x03\x04")

        # check that the file contains the correct data
        with open(filename, "rb") as f:
            data = f.read()

        self.assertEqual(data, b"\x00\x01\x02\x03\x04")

        # clean up
        os.remove(filename)

    def test_no_leak(self):
        filename = tempfile.mktemp()

        # try opening and closing the file a bunch of times
        for _ in range(10000):
            with winunlock_open(filename, "w") as f:
                f.write(self.random_string(100))

        # if we got this far without an error, there was no handle leak

        # clean up
        os.remove(filename)

    def test_read_some_bytes(self):
        # create a temp file with some bytes
        filename = tempfile.mktemp()
        with open(filename, "wb") as f:
            f.write(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09")

        # try reading only some of the bytes with winunlock_open
        with winunlock_open(filename, "rb") as f:
            data = f.read(5)

        self.assertEqual(data, b"\x00\x01\x02\x03\x04")

        # clean up
        os.remove(filename)

    def test_seek_and_tell(self):
        # create a temp file with some bytes
        filename = tempfile.mktemp()
        with open(filename, "wb") as f:
            f.write(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09")

        # try seeking and reading with winunlock_open
        with winunlock_open(filename, "rb") as f:
            f.seek(5)
            data = f.read()

        self.assertEqual(data, b"\x05\x06\x07\x08\x09")

        # try seeking and telling with winunlock_open
        with winunlock_open(filename, "rb") as f:
            f.seek(5)
            position = f.tell()

        self.assertEqual(position, 5)

        # clean up
        os.remove(filename)

    def test_nonexistent_file(self):
        with self.assertRaises(FileNotFoundError):
            with winunlock_open("nonexistent_file.txt", "r") as f:
                pass

    def test_overwriting_content(self):
        filename = tempfile.mktemp()
        with winunlock_open(filename, "w") as f:
            f.write("hello world")
        with winunlock_open(filename, "w") as f:
            f.write("goodbye world")

        with open(filename, "r") as f:
            data = f.read()

        self.assertEqual(data, "goodbye world")

        # clean up
        os.remove(filename)

    def test_appending_content(self):
        filename = tempfile.mktemp()
        with winunlock_open(filename, "w") as f:
            f.write("hello world")
        with winunlock_open(filename, "a") as f:
            f.write(" and goodbye world")

        with open(filename, "r") as f:
            data = f.read()

        self.assertEqual(data, "hello world and goodbye world")

        # clean up
        os.remove(filename)

    def test_large_file(self):
        filename = tempfile.mktemp()
        content = "a" * (1024 * 1024 * 10)  # 10MB

        with winunlock_open(filename, "w") as f:
            f.write(content)

        with winunlock_open(filename, "r") as f:
            data = f.read()

        self.assertEqual(data, content)

        # clean up
        os.remove(filename)

    def test_invalid_parameters(self):
        with self.assertRaises(TypeError):
            with winunlock_open(1234, "r") as f:
                pass
        with self.assertRaises(ValueError):
            with winunlock_open(tempfile.mktemp(), "invalid_mode") as f:
                pass

    def test_read_few_bytes(self):
        # create a temp file with some bytes
        filename = tempfile.mktemp()
        content = "This is a sample string"
        with open(filename, "w") as f:
            f.write(content)

        # Open the file without 'with', read a few bytes, and close
        f = winunlock_open(filename, "r")
        data = f.read(4)  # reads the first 4 bytes/characters
        f.close()

        self.assertEqual(data, content[:4])  # compare with the first 4 characters

        # clean up
        os.remove(filename)


if __name__ == "__main__":
    unittest.main()



from typing import Callable, Union, IO, Any, Iterator
import io
from ._open import open as winunlock_open
from contextlib import contextmanager
import builtins
from os import PathLike
from ._iotype import _IOType

builtins_open = builtins.open


def _open(
    file: Union[PathLike[str], str],
    mode: str = "r",
    buffering: int = -1,
    encoding: Union[str, None] = None,
    errors: Union[str, None] = None,
    newline: Union[str, None] = None,
    closefd: bool = True,
    opener: Union[Callable[[str, int], int], None] = None,
) -> _IOType:
    """
    A helper function that tries Python's built-in open function first and
    falls back to winunlock's open function in case of a PermissionError.

    Args:
        file (str): The path to the file to be opened.
        mode (str, optional): Mode in which the file is opened. Defaults to 'r'.
        buffering (int, optional): Set the buffering policy. Defaults to -1.
        encoding (str, optional): The name of the encoding. Defaults to None.
        errors (str, optional): Specifies how encoding errors are to be handled. Defaults to None.
        newline (str, optional): How newlines mode works. Defaults to None.
        closefd (bool, optional): Close the underlying file descriptor. Defaults to True.
        opener (Callable[[str, int], int], optional): A custom opener. Defaults to None.

    Returns:
        io.TextIOWrapper | io.BufferedRandom | io.BufferedWriter | io.BufferedReader | io.FileIO:
            File object corresponding to the mode.

    """

    # Don't patch if we're not opening a file for reading
    if not "r" in str(mode):
        return builtins_open(
            file, mode, buffering, encoding, errors, newline, closefd, opener
        )

    else:
        try:
            return builtins_open(
                file, mode, buffering, encoding, errors, newline, closefd, opener
            )
        except PermissionError as e:
            try:
                return winunlock_open(
                    file, mode, buffering, encoding, errors, newline, closefd, opener
                )

            # If we can't open the file, raise the original exception instead of the
            # one from open_locked in order for the whole system to behave as if
            # open_locked never existed.
            except Exception as e:
                raise e


@contextmanager
def monkeypatch_open() -> Iterator[None]:
    """
    A context manager that temporarily replaces the built-in open function
    with a custom function that tries Python's built-in open function first and
    falls back to winunlock's open function in case of a PermissionError.

    This context manager is useful when you want to read files that might be
    locked, without modifying the rest of your code.

    Example:
        >>> with monkeypatch_open():
        >>>     with open("locked_file.txt", "r") as f:
        >>>         content = f.read()

    Note:
        The built-in open function will be restored upon exiting the context.

    """
    save_builtins_open = builtins.open
    save_io_open = io.open
    builtins.open = _open  # type: ignore
    io.open = _open  # type: ignore
    yield
    builtins.open = save_builtins_open
    io.open = save_io_open



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="winunlock",
    version="0.1.0",  # Assuming this is version 0.1.0. Update as needed.
    author="Auto Actuary",  # Update with your name or organization name.
    description="A Python module that provides functionality for opening certain types of locked files in Windows.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/autoactuary/winunlock",  # Update with your repo URL.
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # Update with the license you've chosen.
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.6",  # Adjusted to 3.6 for type hinting, but change as needed.
    use_scm_version={
        "write_to": "winunlock/version.py",
    },
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[],
    package_data={
        "": [
            "py.typed",
        ],
    },
)
